//  Binäruhr 2.0
//
//  Copyright (C) 2009,2011  Paul Wilhelm <paul@mosfetkiller.de>
//  http://mosfetkiller.de/?s=binaeruhr2

//  Includes
#define F_CPU 4608000UL //  4,608 MHz
#include <util/atomic.h>
#include <util/delay.h>
#include <avr/interrupt.h>
#include <avr/io.h>
#include <avr/pgmspace.h>

//  Makros
#define set_bit(var, bit) ((var) |= (1 << (bit)))
#define clear_bit(var, bit) ((var) &= (unsigned)~(1 << (bit)))

#define msn(var)  ((var >> 4) & 0xf)
#define lsn(var)  ((var & 0xf))

#define FALSE 0
#define TRUE 1

//  Wochentagsberechnung
#define THIS_CENTURY  20

//  Uhr
#define SECONDS     0
#define MINUTES     1
#define HOURS     2
#define MDAY      3
#define MONTH     4
#define YEAR      5
#define WDAY      6

//  Pinbelegungen Schalter nach GND
#define BTN_port  PORTD
#define BTN_pin   PIND
#define BTN_UP_bit  3   //  Hoch
#define BTN_DN_bit  4   //  Runter
#define BTN_OK_bit  5   //  Bestätigen

#define BTN_PRESS_THRESHOLD 20  //  ms

#define BTN_COUNTER_MAX   100 // ms

//  Charlieplexing
#define CHARLIE_PINS  7
#define CHARLIE_LEDS  35

#define CHARLIE_ddr   DDRB
#define CHARLIE_port  PORTB

#define PLUS(var) ((unsigned)(var))
#define MINUS(var)  ((unsigned)((var) << 4))

//  Verwende Pinnummerierung an Charlieplexing Template (1..7)
#define CH_SECONDS_0  PLUS(7) | MINUS(3)
#define CH_SECONDS_1  PLUS(2) | MINUS(5)
#define CH_SECONDS_2  PLUS(4) | MINUS(6)
#define CH_SECONDS_3  PLUS(3) | MINUS(5)
#define CH_SECONDS_4  PLUS(7) | MINUS(6)
#define CH_SECONDS_5  PLUS(4) | MINUS(3)

#define CH_MINUTES_0  PLUS(1) | MINUS(3)
#define CH_MINUTES_1  PLUS(3) | MINUS(4)
#define CH_MINUTES_2  PLUS(5) | MINUS(6)
#define CH_MINUTES_3  PLUS(5) | MINUS(4)
#define CH_MINUTES_4  PLUS(2) | MINUS(1)
#define CH_MINUTES_5  PLUS(1) | MINUS(2)

#define CH_HOURS_0  PLUS(2) | MINUS(4)
#define CH_HOURS_1  PLUS(6) | MINUS(5)
#define CH_HOURS_2  PLUS(3) | MINUS(2)
#define CH_HOURS_3  PLUS(2) | MINUS(3)
#define CH_HOURS_4  PLUS(4) | MINUS(5)

#define CH_WDAY_0 PLUS(4) | MINUS(2)
#define CH_WDAY_1 PLUS(3) | MINUS(1)
#define CH_WDAY_2 PLUS(6) | MINUS(7)

#define CH_MDAY_0 PLUS(5) | MINUS(3)
#define CH_MDAY_1 PLUS(6) | MINUS(4)
#define CH_MDAY_2 PLUS(7) | MINUS(5)
#define CH_MDAY_3 PLUS(5) | MINUS(7)
#define CH_MDAY_4 PLUS(4) | MINUS(1)

#define CH_MONTH_0  PLUS(1) | MINUS(4)
#define CH_MONTH_1  PLUS(3) | MINUS(6)
#define CH_MONTH_2  PLUS(1) | MINUS(5)
#define CH_MONTH_3  PLUS(5) | MINUS(1)

#define CH_YEAR_0 PLUS(5) | MINUS(2)
#define CH_YEAR_1 PLUS(6) | MINUS(3)
#define CH_YEAR_2 PLUS(4) | MINUS(7)
#define CH_YEAR_3 PLUS(2) | MINUS(6)
#define CH_YEAR_4 PLUS(7) | MINUS(4)
#define CH_YEAR_5 PLUS(3) | MINUS(7)

#define PWM_COUNTER_MAX 20

//  Autodim
#define DIM_EVENING 22
#define DIM_MORNING 6

//  Variablen
volatile unsigned char clock[7] = { 0, 0, 0, 1, 1, 0, 0 };
volatile unsigned char clock_min[6] = {  0,  0,  0,  1,  1,  0 };
volatile unsigned char clock_max[6] = { 59, 59, 23, 31, 12, 31 };

volatile unsigned char clock_running = FALSE, new_year = TRUE, new_day = TRUE, new_minute = TRUE;

volatile unsigned char mday_max[12] = { 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 };
volatile unsigned char month_table[12] PROGMEM = { 0, 3, 3, 6, 1, 4, 6, 2, 5, 0, 3, 5 };

volatile unsigned int btn_up_counter = 0, btn_dn_counter = 0, btn_ok_counter = 0;

volatile unsigned char brightness = 4;

volatile unsigned char brightness_table[5] = { 1, 2, 5, 10, 20 };

//  [Parameter][Bit]
volatile unsigned char charlie_pins[7][6] PROGMEM =
{
  { CH_SECONDS_0, CH_SECONDS_1, CH_SECONDS_2, CH_SECONDS_3, CH_SECONDS_4, CH_SECONDS_5 },
  { CH_MINUTES_0, CH_MINUTES_1, CH_MINUTES_2, CH_MINUTES_3, CH_MINUTES_4, CH_MINUTES_5 },
  { CH_HOURS_0, CH_HOURS_1, CH_HOURS_2, CH_HOURS_3, CH_HOURS_4, 0 },
  { CH_MDAY_0, CH_MDAY_1, CH_MDAY_2, CH_MDAY_3, CH_MDAY_4, 0 },
  { CH_MONTH_0, CH_MONTH_1, CH_MONTH_2, CH_MONTH_3, 0, 0 },
  { CH_YEAR_0, CH_YEAR_1, CH_YEAR_2, CH_YEAR_3, CH_YEAR_4, CH_YEAR_5 },
  { CH_WDAY_0, CH_WDAY_1, CH_WDAY_2, 0, 0, 0 },
};

volatile unsigned char bit_max[7] PROGMEM = { 5, 5, 4, 4, 3, 5, 2 };

//  Prototypen
void init();
void recalc_mday_max();
void check_date();
void recalc_wday();
void charlie_set_pixel(unsigned char pin_pair);


//  Main
int main(void)
{
  //  Initialisieren
  init();


  //  Zeit und Datum setzen
  unsigned char parameter_this = 0; //  Sekunden

  char set = FALSE, changed_this = TRUE, next_parameter = TRUE;

  while (!set)
  {
    //  Monatstagetabellen und Wochentag bei geändertem Parameterwert neu berechnen, Datum überprüfen
    if (changed_this)
    {
      changed_this = FALSE;

      recalc_mday_max();
      check_date();
      recalc_wday();
    }

    if (next_parameter)
    {
      next_parameter = FALSE;

      //  Aktuellen Parameter aufblinken lassen
      unsigned char temp = clock[parameter_this];
      for (unsigned char i = 0; i < 3; i++)
      {
        clock[parameter_this] = 0;
        _delay_ms(10);
        clock[parameter_this] = 0xff;
        _delay_ms(10);
      }
      clock[parameter_this] = temp;
    }


    //  BTN_UP: Aktuellen Parameterwert inkrementieren
    if (btn_up_counter >= BTN_PRESS_THRESHOLD && btn_dn_counter == 0)
    {
      if (clock[parameter_this] < clock_max[parameter_this])
      {
        clock[parameter_this]++;
      }
      else
      {
        clock[parameter_this] = clock_min[parameter_this];
      }

      while (btn_up_counter >= BTN_PRESS_THRESHOLD);
      btn_up_counter = 0;

      changed_this = TRUE;
    }

    //  BTN_DN: Aktuellen Parameterwert dekrementieren
    if (btn_dn_counter >= BTN_PRESS_THRESHOLD && btn_up_counter == 0)
    {
      if (clock[parameter_this] > clock_min[parameter_this])
      {
        clock[parameter_this]--;
      }
      else
      {
        clock[parameter_this] = clock_max[parameter_this];
      }

      while (btn_dn_counter >= BTN_PRESS_THRESHOLD);
      btn_dn_counter = 0;

      changed_this = TRUE;
    }

    //  BTN_OK: Nächsten Parameter auswählen
    if (btn_ok_counter >= BTN_PRESS_THRESHOLD)
    {
      if (parameter_this < 5) parameter_this++; else parameter_this = 0;  //  Bis YEAR, dann wieder bei SECONDS angefangen

      while (btn_ok_counter >= BTN_PRESS_THRESHOLD);
      btn_ok_counter = 0;

      next_parameter = TRUE;
    }

    //  BTN_DN + DTN_UP: Menü beenden
    if (btn_up_counter >= BTN_PRESS_THRESHOLD && btn_dn_counter >= BTN_PRESS_THRESHOLD)
    {
      set = TRUE;
    }
  }

  //  Alle Tastenzähler resetten
  btn_up_counter = 0;
  btn_dn_counter = 0;
  btn_ok_counter = 0;

  //  Normaler Uhrenbetrieb
  clock_running = TRUE;

  while (1)
  {
    //  Bei Jahreswechsel auf Schaltjahr prüfen und ggf. Tabellen aktualisieren
    if (new_year)
    {
      new_year = FALSE;
      recalc_mday_max();
    }

    //  Bei Tageswechsel Wochentag neu berechnen
    if (new_day)
    {
      new_day = FALSE;
      recalc_wday();
    }


    //  BTN_UP: Helligkeit erhöhen
    if (btn_up_counter >= BTN_PRESS_THRESHOLD)
    {
      if (brightness < 4) brightness++;
      while (btn_up_counter >= BTN_PRESS_THRESHOLD);
      btn_up_counter = 0;
    }

    //  BTN_DN: Helligkeit erniedrigen
    if (btn_dn_counter >= BTN_PRESS_THRESHOLD)
    {
      if (brightness > 0) brightness--;
      while (btn_dn_counter >= BTN_PRESS_THRESHOLD);
      btn_dn_counter = 0;
    }


    //  Autodim
    if (new_minute)
    {
      new_minute = FALSE;

      if (clock[HOURS] == DIM_EVENING && clock[MINUTES] == 0)
      {
        brightness = 0;
      }

      if (clock[HOURS] == DIM_MORNING && clock[MINUTES] == 0)
      {
        brightness = 2;
      }
    }
  }
}


//  Initialisieren
void init()
{
  //  Ports vorbereiten

  //  Button Hoch, Button Runter, Button Bestätigen (Input, H (PullUp))
  BTN_port |= (1 << BTN_UP_bit) | (1 << BTN_DN_bit) | (1 << BTN_OK_bit);

  //  TIMER0 vorbereiten (Zeit, Tastenzähler)
  TCCR0A = (1 << WGM01);        //  CTC Mode
  TCCR0B = (1 << CS01) | (1 << CS00); //  Prescaler 64
  OCR0A = 71;             //  Compare A
  set_bit(TIMSK, OCIE0A);

  //  TIMER1 vorbereiten (Charlieplexing)
  TCCR1B = (1 << WGM12) | (1 << CS10);  //  CTC Mode, Prescaler 1
  OCR1AH = 0;               //  Compare A
  OCR1AL = 180;
  set_bit(TIMSK, OCIE1A);

  sei();
}


//  Auf Schaltjahr prüfen und ggf. Tabellen aktualisieren
void recalc_mday_max()
{
  if ((clock[YEAR] % 4 == 0) && (clock[YEAR] % 400) != 0)
  {
    mday_max[1] = 29; //  Februar hat 29 Tage
  }
  else
  {
    mday_max[1] = 28; //  Februar hat 28 Tage
  }

  clock_max[MDAY] = mday_max[clock[MONTH] - 1];
}


//  Datum auf Validität prüfen
void check_date()
{
  //  Hat der aktuelle Monat weniger Tage als in MDAY angegeben, nach unten korrigieren
  if (clock[MDAY] > clock_max[MDAY]) clock[MDAY] = clock_max[MDAY];
}


//  Wochentag neu berechnen
void recalc_wday()
{
  clock[WDAY] = (clock[MDAY] % 7) + pgm_read_byte(&month_table[clock[MONTH] - 1]) + ((clock[YEAR] + clock[YEAR] / 4) % 7) + (3 - (THIS_CENTURY % 4)) * 2;
  if (clock[MONTH] < 3 && (clock[YEAR] % 4 == 0)) clock[WDAY]--;
  clock[WDAY] %= 7;
  if (clock[WDAY] == 0) clock[WDAY] = 7;
}


//  Charlieplexing: Einzelne LED einschalten, Rest ausschalten / Alle LEDs ausschalten
void charlie_set_pixel(unsigned char pin_pair)
{
  //  Alle LEDs ausschalten
  if (pin_pair == 0)
  {
    CHARLIE_ddr = 0;  //  Alle Input
    CHARLIE_port = 0; //  Alle Z

    return;
  }

  //  Einzelne LED einschalten
  unsigned char pin1 = lsn(pin_pair) - 1;
  unsigned char pin2 = msn(pin_pair) - 1;

  //  Die Reihenfolge ist wichtig, ansonsten leuchten manche LEDs unerwünscht
  CHARLIE_ddr = (1 << pin1) | (1 << pin2);  //  Pin 1 (Anode) Output; Pin 2 (Kathode) Output
  CHARLIE_port = (1 << pin1);         //  Pin 1 (Anode) H;    Pin 2 (Kathode) L
}


//  TIMER0 (Zeit, Tastenzähler)
SIGNAL(SIG_OUTPUT_COMPARE0A)
{
  static unsigned int mseconds;

  if (clock_running)
  {
    if (mseconds < 999)
    {
      mseconds++;
    }
    else
    {
      mseconds = 0;

      if (clock[SECONDS] < clock_max[SECONDS])
      {
        clock[SECONDS]++;
      }
      else
      {
        clock[SECONDS] = clock_min[SECONDS];

        new_minute = TRUE;

        if (clock[MINUTES] < clock_max[MINUTES])
        {
          clock[MINUTES]++;
        }
        else
        {
          clock[MINUTES] = clock_min[MINUTES];

          if (clock[HOURS] < clock_max[HOURS])
          {
            clock[HOURS]++;
          }
          else
          {
            clock[HOURS] = clock_min[HOURS];

            new_day = TRUE;

            if (clock[MDAY] < clock_max[MDAY])
            {
              clock[MDAY]++;
            }
            else
            {
              clock[MDAY] = clock_min[MDAY];

              if (clock[MONTH] < clock_max[MONTH])
              {
                clock[MONTH]++;
              }
              else
              {
                clock[MONTH] = clock_min[MONTH];

                if (clock[YEAR] < clock_max[YEAR])  //  Nur 6 LEDs Jahresanzeige
                {
                  clock[YEAR]++;
                }
                else
                {
                  clock[YEAR] = clock_min[YEAR];
                }

                new_year = TRUE;
              }
            }
          }
        }
      }
    }
  }

  //  Tastenzähler

  //  Hoch
  if (bit_is_clear(BTN_pin, BTN_UP_bit))
  {
    if (btn_up_counter < BTN_COUNTER_MAX) btn_up_counter++;
  }
  else
  {
    if (btn_up_counter > 0) btn_up_counter--;
  }

  //  Runter
  if (bit_is_clear(BTN_pin, BTN_DN_bit))
  {
    if (btn_dn_counter < BTN_COUNTER_MAX) btn_dn_counter++;
  }
  else
  {
    if (btn_dn_counter > 0) btn_dn_counter--;
  }

  //  Bestätigen
  if (bit_is_clear(BTN_pin, BTN_OK_bit))
  {
    if (btn_ok_counter < BTN_COUNTER_MAX) btn_ok_counter++;
  }
  else
  {
    if (btn_ok_counter > 0) btn_ok_counter--;
  }
}


//  TIMER1 (Charlieplexing)
SIGNAL(SIG_OUTPUT_COMPARE1A)
{
  static unsigned char parameter_this, bit_this;
  static unsigned char pwm_counter;

  if (brightness_table[brightness] > pwm_counter && bit_is_set(clock[parameter_this], bit_this))
  {
    charlie_set_pixel(pgm_read_byte(&charlie_pins[parameter_this][bit_this]));
  }
  else
  {
    charlie_set_pixel(0);
  }

  //  Nächste LED (Bits und Parameter durchlaufen)
  if (bit_this < pgm_read_byte(&bit_max[parameter_this]))
  {
    bit_this++;
  }
  else
  {
    bit_this = 0;

    if (parameter_this < 6)
    {
      parameter_this++;
    }
    else
    {
      parameter_this = 0;

      if (pwm_counter < PWM_COUNTER_MAX) pwm_counter++; else pwm_counter = 0;
    }
  }
}
